from typing import TYPE_CHECKING, Any

from langchain._api import create_importer

if TYPE_CHECKING:
    from langchain_community.tools import (
        BaseSparkSQLTool,
        InfoSparkSQLTool,
        ListSparkSQLTool,
        QueryCheckerTool,
        QuerySparkSQLTool,
    )

# Create a way to dynamically look up deprecated imports.
# Used to consolidate logic for raising deprecation warnings and
# handling optional imports.
DEPRECATED_LOOKUP = {
    "BaseSparkSQLTool": "langchain_community.tools",
    "QuerySparkSQLTool": "langchain_community.tools",
    "InfoSparkSQLTool": "langchain_community.tools",
    "ListSparkSQLTool": "langchain_community.tools",
    "QueryCheckerTool": "langchain_community.tools",
}

_import_attribute = create_importer(__package__, deprecated_lookups=DEPRECATED_LOOKUP)


def __getattr__(name: str) -> Any:
    """Look up attributes dynamically."""
    return _import_attribute(name)


__all__ = [
    "BaseSparkSQLTool",
    "QuerySparkSQLTool",
    "InfoSparkSQLTool",
    "ListSparkSQLTool",
    "QueryCheckerTool",
]
